#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (C) 2014 Canonical
#
# Authors:
#  Didier Roche
#  Łukasz 'sil2100' Zemczak
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""CI Train Build Job

Prepare, build and check packages.

Environment variables:

PACKAGES_TO_REBUILD
    Only rebuild the specified source package names (space separated).

WATCH_ONLY
    Don't trigger a build, only watch what's already building in the PPA.

FORCE_REBUILD
    Force rebuilding an MP even if there is no diff with dest.

IGNORE_STEP
    Perform build even if one was already done.

IGNORE_MISSING_TWINS
    Don't enforce building a -gles twin for packages that usually have it.

DONT_REORDER_PREREQS
    Don't forcefully reorder branches if MP prerequisites are used.

TAKE_WHOLE_COMMIT_MESSAGE
    Use whole commit message in changelog, not just the brief part.

DO_NOT_APPEND_RTM_TO_VERSION
    Do not rewrite the package version when doing uploads to ubuntu-rtm
    (e.g. do not add the ~rtm to the upstream part)

DEBUG
    Enable debug infos
"""

import glob
import lazr
import logging
import os
import re
import shutil
import subprocess
import sys

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from citrain import watch_ppa
from cupstream2distro import (
    branchhandling,
    launchpadmanager,
    packagemanager,
    silomanager,
    tools,
)
from cupstream2distro.settings import (
    BOT_DEBFULLNAME,
    PROJECT_CONFIG_SUFFIX,
    SERIES_VERSION,
    SILO_EMPTY,
    SILOS_DIR,
)
from cupstream2distro.tools import clean_source
from cupstream2distro.utils import env

from pprint import pformat


def main():
    logging.basicConfig(
        level=logging.DEBUG if env.DEBUG == 'true' else logging.INFO,
        format="%(asctime)s %(levelname)s %(message)s")

    logging.debug(str(env))
    silo_config = silomanager.load_config()
    logging.debug('Starting with this config:' + pformat(silo_config))
    silomanager.set_config_status(
        silo_config,
        silomanager.SILO_STATE_BUILDING,
        "Preparing packages",
        ping=False)

    src_ppa = launchpadmanager.get_resource_from_token(
        silo_config["global"]["ppa"])
    dest = launchpadmanager.get_resource_from_token(
        silo_config["global"]["dest"])
    series = launchpadmanager.get_resource_from_token(
        silo_config["global"]["series"])

    sync_request = False

    if "source_archive" in silo_config:
        logging.debug(
            "A sync request has been requested. "
            "Is landing sync from silo? {}".format(
                "source_sync_from_silo" in silo_config))
        sync_request = True

    if not env.PACKAGES_TO_REBUILD and env.WATCH_ONLY != 'true':
        if silomanager.get_config_step(silo_config) != SILO_EMPTY:
            if "source_sync_from_silo" in silo_config:
                logging.warning(
                    "A landing sync is configured here, "
                    "doing a whole rebuild.")
            elif env.IGNORE_STEP == 'true':
                logging.warning(
                    "You already tried to build everything. "
                    "But the ignore option is set.")
            else:
                logging.error(
                    "You already tried to build everything. You need to "
                    "either:\n - use the ignore step option (to rebuild all "
                    "MPs)\n - use the \"only rebuild packages\" option (to "
                    "rebuild only some components associated to a MP) \n - "
                    "use the watch only option (to only check ppa status "
                    "without any rebuild).")
                silomanager.set_config_status(
                    silo_config,
                    silomanager.SILO_STATE_BUILD_FAILED,
                    "Can't build: prevent a whole rebuild as one full build "
                    "was already successful. Is that on purpose? Please "
                    "check available options.")
                return 1
        # remove everything but the archived .project_ files
        logging.info(
            "Whole rebuild, ensuring that the silo is in a coherent state. "
            "Keep the published files")
        for filename in os.listdir("."):
            # don't remove archive file and config file
            if (not ".{}_".format(PROJECT_CONFIG_SUFFIX) in filename
                    and "config" not in filename):
                try:
                    shutil.rmtree(filename)
                except OSError:
                    os.remove(filename)

    # reset the status
    if not silomanager.set_config_step(silo_config, SILO_EMPTY):
        logging.error("Couldn't reset the step in the configuration file.")
        silomanager.set_config_status(
            silo_config,
            silomanager.SILO_STATE_BUILD_FAILED,
            "Build failed: couldn't reset status or parameters for job.")
        return 1

    # detect if we have a prepare only source not in configuration
    components_not_in_config = []
    for source in env.PACKAGES_TO_REBUILD.split():
        if source not in silo_config["mps"] and (
                not sync_request or source not in silo_config["sources"]):
            components_not_in_config.append(source)
    if components_not_in_config:
        message = (
            "One or more components provided in rebuildpackages ({}) are "
            "not in the MP silo configuration. Ask for a reconfiguration of "
            "the silo with the new MPs to add please.".format(
                ", ".join(components_not_in_config)))
        logging.error(message)
        silomanager.set_config_status(
            silo_config,
            silomanager.SILO_STATE_BUILD_FAILED,
            "Can't build: " + message)
        return 1

    # cycle through all branches and prepare them
    components_to_prepare = {}
    for source in silo_config["mps"]:
        if ((not env.PACKAGES_TO_REBUILD and env.WATCH_ONLY != 'true') or
                source in env.PACKAGES_TO_REBUILD.split()):
            logging.info("Adding {} MP(s) to prepare".format(source))
            for mp_link in silo_config['mps'][source]:
                components_to_prepare.setdefault(source, []).append(
                    launchpadmanager.get_resource_from_token(mp_link))
            # cleaning (in case of partial rebuild)
            clean_source(source)

    all_sources = components_to_prepare.keys() + silo_config["sources"]
    logging.debug('all_sources: ' + pformat(all_sources))
    missing = packagemanager.check_if_packages_require_twin_upload(all_sources)
    if missing:
        message = (
            "Some projects are missing their 'twin package' uploads (e.g. "
            "their -gles counter-parts): {}.".format(", ".join(missing)))
        if env.IGNORE_MISSING_TWINS == 'true':
            logging.warning(message + " But the ignore flag was set.")
        else:
            logging.error(message)
            silomanager.set_config_status(
                silo_config,
                silomanager.SILO_STATE_BUILD_FAILED,
                "Can't build: " + message)
            return 1

    # prepare source and dput by one
    src_ppa_name = "{}/{}/{}".format(
        src_ppa.owner.name, src_ppa.distribution.name, src_ppa.name)
    dest_ppa_name = "{}/{}/{}".format(
        dest.owner.name, src_ppa.distribution.name, dest.name)
    logging.debug('src_ppa_name: ' + src_ppa_name)
    logging.debug('dest_ppa_name: ' + dest_ppa_name)

    src_pkgs_to_upload = {}
    for source_package_name in components_to_prepare:
        logging.info(
            "=== Trying to prepare {} ===".format(source_package_name))

        # get dest branch first
        dest_branch = components_to_prepare[
            source_package_name][0].target_branch.web_link
        logging.info("Branching {} to {}".format(
            dest_branch, source_package_name))
        branchhandling.get_branch(dest_branch, source_package_name)
        bzr_pending_branch_uri = os.path.abspath(source_package_name)
        os.chdir(bzr_pending_branch_uri)
        dest_tip_bzr_rev = branchhandling.get_tip_bzr_revision()
        dest_version = packagemanager.get_packaging_version()

        # first, try looking into the merge list and re-order it if there are
        # any prereqs present if no prereqs are present, the components list
        # stays the same
        if env.DONT_REORDER_PREREQS != 'true':
            components = tools.reorder_branches_regarding_prereqs(
                components_to_prepare[source_package_name])
        else:
            components = components_to_prepare[source_package_name]

        # get all branches one after another and merge them in
        # bzr_pending_branch_uri
        for mp in components:
            branch = mp.source_branch.web_link
            logging.info("Trying to merge {} (at rev {}) into {}".format(
                branch,
                mp.source_branch.revision_count,
                mp.target_branch.web_link))

            bugs_fixes = ""
            for bug in mp.getRelatedBugTasks():
                if not bugs_fixes:
                    bugs_fixes = "Fixes: "
                else:
                    bugs_fixes += ", "
                bugs_fixes += bug.self_link.split('/')[-1]

            # use a set to fetch the reviewer list
            approved_by_set = set()
            for review in mp.all_comments:
                if review.vote == "Approve":
                    approved_by_set.add(review.author.display_name)
                elif review.vote in ("Disapprove", "Needs Fixing"):
                    if review.author.display_name in approved_by_set:
                        approved_by_set.remove(review.author.display_name)

            approved_by = ""
            for reviewer in approved_by_set:
                if not approved_by:
                    approved_by = "\nApproved by: "
                else:
                    approved_by += ", "
                approved_by += reviewer

            commit_message = mp.commit_message
            if commit_message:
                commit_message = "{} {}{}".format(
                    commit_message.encode('utf-8'),
                    bugs_fixes,
                    approved_by.encode('utf-8'))

            # grab committers
            committers = branchhandling.grab_committers_compared_to(
                bzr_pending_branch_uri, branch)

            try:
                if not branchhandling.merge_branch(
                        bzr_pending_branch_uri,
                        branch,
                        commit_message,
                        authors=committers):
                    logging.error(
                        "Conflicts when attempting to merge {} into previous "
                        "merged + {}. You will need to reconcile manually "
                        "branches that you provided as a parameter for "
                        "merging into the destination branch and rerun this "
                        "job".format(branch, mp.target_branch.web_link))
                    os.chdir('..')
                    silomanager.set_config_status(
                        silo_config,
                        silomanager.SILO_STATE_BUILD_FAILED,
                        "Can't build: Merge conflict.")
                    return 1
            # TODO: we should loop to show if there is more issues and then
            # exit
            except branchhandling.NoCommitFoundException:
                message = (
                    "There is no commit message in {} and we can't get it "
                    "from debian/changelog. Please check that you set a "
                    "commit message on all your MPs if you didn't touch "
                    "debian/changelog.".format(branch))
                logging.error(message)
                silomanager.set_config_status(
                    silo_config,
                    silomanager.SILO_STATE_BUILD_FAILED,
                    "Can't build: " + message)
                return 1

        #  Get changelog informations
        logging.info("Get changelog info")
        prep_version = packagemanager.get_packaging_version(
        )

        # Check version in ppa
        logging.info("Get previous version in silo")
        version_in_ppa = packagemanager.get_current_version_for_series(
            source_package_name, series.name, src_ppa_name)
        if packagemanager.is_version1_higher_than_version2(
                version_in_ppa,
                prep_version):
            logging.info(
                "A version in the ppa ({}) is higher than the proposed "
                "version in bzr ({}) (previous tests/builds failing?). "
                "Basing on that one.".format(
                    version_in_ppa,
                    dest_version))
            previous_packaging_version = version_in_ppa
        else:
            previous_packaging_version = prep_version

        # Check version in distro
        logging.info("Get version available in distro")
        distro_version = packagemanager.get_current_version_for_series(
            source_package_name,
            series.name,
            dest=dest.distribution.main_archive)

        # if the target is the distro, we ensure we have all releases in
        # proposed branch
        if launchpadmanager.is_dest_distro_archive(
                silo_config["global"]["dest"]):
            prev_version = distro_version
        # target is a ppa
        else:
            prev_version = packagemanager.get_current_version_for_series(
                source_package_name, series.name, dest=dest)

        # ensure that the previous release (in distro or in dest ppa) ended up
        # in the vcs
        force_rebuild = False
        if env.FORCE_REBUILD == 'true':
            logging.info(
                "Force rebuild option set. Will update even if destination "
                "don't contain latest version requested.")
            force_rebuild = True
        else:
            with open("debian/changelog") as changelog:
                if not packagemanager.is_version_in_changelog(
                        prev_version,
                        changelog):
                    message = (
                        "A version ({}) is available at the destination "
                        "archive for that component but is not in the "
                        "destination branch which is still at {}. You "
                        "need to ensure that your version contains the "
                        "fix in the destination or you can force "
                        "rebuild to bypass the check.".format(
                            prev_version,
                            prep_version))
                    logging.warning(message)
                    os.chdir('..')
                    silomanager.set_config_status(
                        silo_config,
                        silomanager.SILO_STATE_BUILD_FAILED,
                        "Can't build: Some projects are missing "
                        "revisions from distribution.")
                    return 1

        if not launchpadmanager.is_dest_distro_archive(
                silo_config["global"]["dest"]):
            # if the version in the distro is higher than the previous
            # packaging version (either the version in intermediate ppa or
            # packaging bzr version We are going to force releasing, but basing
            # on this version
            if packagemanager.is_version1_higher_than_version2(
                    distro_version,
                    previous_packaging_version):
                logging.info(
                    "The version in distro ({}) is higher than what is "
                    "proposed in the bzr version for that branch ({}). "
                    "Basing on the distro version.".format(
                        distro_version, previous_packaging_version))
                logging.info(
                    "Forcing release mode as the version in distro ({}) is "
                    "higher than current destination ppa ({})".format(
                        distro_version, previous_packaging_version))
                previous_packaging_version = distro_version
                force_rebuild = True

        # Check if we need to republish
        dest_source_package = None
        logging.info("Check if we need to release a new package")
        os.chdir('..')
        dest_source_package = packagemanager.get_source_package_from_dest(
            source_package_name,
            dest,
            prev_version,
            series.name)
        if dest_source_package:
            dest_source_package = os.path.abspath(dest_source_package)
        os.chdir(bzr_pending_branch_uri)
        # nothing to release
        f = packagemanager.is_new_content_relevant_since_old_published_source
        if not f(dest_source_package):
            if not force_rebuild:
                logging.info(
                    "No new useful revision published compared to "
                    "dest, no need to upload this component")
                os.chdir('..')
                continue
            logging.info(
                "No new useful revision published but force rebuilding is set")

        if launchpadmanager.is_dest_distro_archive(
                silo_config["global"]["dest"]):
            # only version if we don't see the flag to not version it.
            if not packagemanager.has_dont_change_version_flag():
                version = packagemanager.create_new_packaging_version(
                    previous_packaging_version,
                    series_version=SERIES_VERSION[series.name],
                    do_not_append_tag=(
                        env.DO_NOT_APPEND_RTM_TO_VERSION == "true"))
            else:
                version = packagemanager.get_packaging_version()
        else:
            version = packagemanager.create_new_packaging_version(
                previous_packaging_version,
                destppa=dest_ppa_name,
                series_version=SERIES_VERSION[series.name],
                do_not_append_tag=(
                    env.DO_NOT_APPEND_RTM_TO_VERSION == "true"))

        # Refresh symbol files
        logging.info("Update symbol files version if needed")
        packagemanager.refresh_symbol_files(version)

        logging.info("Prepare changelog content")
        authors_commits = {}

        try:
            # first try: in case someone release and release a version without
            # merging back (we don't want to duplicate the commits). This is
            # the main case anyway.
            log_content = branchhandling.log_diff_from_tag(
                dest_version)
        except:
            try:
                # second hammer: we then recollect from the last version in the
                # destination (some lines can be duplicated)
                log_content = branchhandling.log_diff_from_tag(
                    prev_version)
            except:
                # last hammer with FIXME: -100 is a hack as we don't have
                # anymore any information on when the last release happen.
                # People are transitionned to a tag system.
                log_content = branchhandling.log_diff_since_last_release(
                    branchhandling.log_diff("-100"))
        alreadyfixed_bugs = set()
        with open("debian/changelog") as f:
            alreadyfixed_bugs = packagemanager.collect_bugs(
                f, source_package_name)
        (authors_commits, bugs_set) = branchhandling.collect_author_commits(
            log_content,
            alreadyfixed_bugs,
            take_whole_commit=env.TAKE_WHOLE_COMMIT_MESSAGE == 'true')
        if force_rebuild:
            authors_commits.setdefault(
                BOT_DEBFULLNAME, []).append("New rebuild forced")
        packagemanager.update_changelog(
            version, series.name, None, authors_commits)
        branchhandling.commit_release(version)

        logging.info("Sync launchpad with current proposed work")
        launchpadmanager.open_bugs_for_source(
            bugs_set, source_package_name, series.name, dest.distribution.name)

        logging.info("Build source package")
        packagemanager.build_source_package(
            series.name, prev_version)
        os.chdir('..')
        newdsc_path = "{}_{}.dsc".format(
            source_package_name, version.split(':')[-1])

        # second diff step: in case the source package was created from nothing
        # (like a diff between trunk and the source package, but none after
        # newly .dsc created extraction)
        if not force_rebuild and dest_source_package:
            logging.info(
                "Check that the newly created source "
                "package has relevant diff")
            if not packagemanager.is_relevant(
                    newdsc_path, dest_source_package):
                logging.warning(
                    "There was a diff between trunk and the source "
                    "package in ubuntu triggering a source package build. "
                    "However, after that one was created, we diffed again "
                    "between that source in ubuntu and the newly created "
                    "source extracted and found nothing relevant. Not upload "
                    "that one then. Please check that the relevant diff in "
                    "the package creation is justified by diffing manually "
                    "trunk and the source in ubuntu.")

        # check the .changes file: that we don't have an empty changelog
        logging.info("Checking finale changelog")
        with open("{}_{}_source.changes".format(
                source_package_name,
                version.split(':')[-1])) as f:
            for line in f:
                if re.match("   \* *$", line):
                    message = (
                        "You generated an emtpy changelog content for {}. "
                        "This generally means that one or many merge "
                        "proposals have a change in debian/changelog with "
                        "no UNRELEASED or no content in the changelog. "
                        "If you modify debian/changelog in your commit, "
                        "the commit message for the merge proposal isn't "
                        "used, so you need to put yourself some changelog "
                        "content into that file. Please change those MP "
                        "and rerun the build job.".format(source_package_name))
                    logging.error(message)
                    silomanager.set_config_status(
                        silo_config,
                        silomanager.SILO_STATE_BUILD_FAILED,
                        "Build failed: Some packages generate an empty "
                        "changelog. Check the job for more info.")
                    return 1

        # Prepare if a packaging change is part of the diff
        logging.info("Generate a diff if meaningfull packaging changes")
        diff_filepath = tools.get_packaging_diff_filename(
            source_package_name, version)
        olddsc_path = None
        if dest_source_package:
            for file in os.listdir(os.path.dirname(dest_source_package)):
                if file.endswith(".dsc"):
                    olddsc_path = os.path.join(
                        os.path.dirname(dest_source_package), file)
        packagemanager.generate_diff_between_dsc(
            diff_filepath, olddsc_path, newdsc_path)
        # Save infos for uploading later on.
        src_pkgs_to_upload[source_package_name] = (
            version,
            branch,
            dest_tip_bzr_rev,
            prev_version,
            ".")

    # this landing has a sync request in it
    if sync_request and env.WATCH_ONLY != 'true':
        source_archive = launchpadmanager.get_resource_from_token(
            silo_config['source_archive'])
        source_series = silo_config["source_series"]

        if "source_sync_from_silo" in silo_config:
            logging.debug("Syncing whole silo contents from the selected silo")

            request_id = silo_config["source_sync_from_silo"]
            if request_id.isdigit():
                # check if the silo still exists
                source_distro = source_archive.distribution.name
                silo_name = silo_config["source_archive"].split("/")[-1]
                source_silo_config = silomanager.load_config(
                    os.path.join(
                        SILOS_DIR, source_distro, silo_name))
                if (not source_silo_config or
                        source_silo_config["requestid"] != request_id or
                        source_silo_config["global"]["status"]["state"] ==
                        silomanager.SILO_STATE_CLEANING):
                    logging.info(
                        "Warning! /!\ The source silo has been removed, "
                        "retargetting to the {} archive instead".format(
                            source_distro))
                    source_archive = launchpadmanager.get_distribution_archive(
                        source_distro)
                    silo_config["source_archive"] = source_archive.self_link
                    del silo_config["source_sync_from_silo"]
                elif "source_sync_all_contents" in silo_config:
                    # in case where the silo still exists, take the most
                    # up-to-date list of packages
                    source_series_object = launchpadmanager.get_series(
                        source_series,
                        distribution=source_archive.distribution.name)
                    sources_to_sync = [i.source_package_name for i in
                                       source_archive.getPublishedSources(
                                           distro_series=source_series_object,
                                           status="Published")]

                    if not sources_to_sync:
                        logging.error(
                            "The silo sync generates an empty list of packages"
                            " - please make sure the sync is fetching sources "
                            "from the right place. Contact the landing team "
                            "in case of any inconsistencies.")
                        silomanager.set_config_status(
                            silo_config,
                            silomanager.SILO_STATE_BUILD_FAILED,
                            "Build failed: silo sync generates an empty "
                            "sources list. Check configuration.")
                        return 1

                    silo_config["sources"] = sources_to_sync
                    silomanager.save_config(silo_config)

        if not env.PACKAGES_TO_REBUILD:
            sources_to_sync = silo_config["sources"]
        else:
            sources_to_sync = [pkg for pkg in silo_config[
                "sources"] if pkg in env.PACKAGES_TO_REBUILD.split()]

        main_dir = os.getcwd()
        logging.debug("The current main directory: {}".format(main_dir))

        # XXX: we might not need build/ anymore - check that
        build_dir = os.path.join(main_dir, "build")
        if sources_to_sync and os.path.isdir(build_dir):
            # if we're doing rebuilds, clean the build directory
            logging.debug("Removing the build directory: {}".format(build_dir))
            shutil.rmtree(build_dir)

        for pkg in sources_to_sync:
            logging.info("Looking at {} as per sync request".format(pkg))
            current_version = packagemanager.get_current_version_for_series(
                pkg,
                source_series,
                dest=source_archive)

            # prepare new version number
            # TODO: handle the case of direct-uploads that do not bump the
            #  upstream version, only the ubuntu number.
            #  The ~rtm bits need to be then added after the ubuntu bits then
            new_version = current_version
            if env.DO_NOT_APPEND_RTM_TO_VERSION != "true":
                if (source_archive.distribution.name == "ubuntu" and
                        dest.distribution.name == "ubuntu-rtm"):
                    logging.debug("Rewriting the version to: {}".format(
                        new_version))
                    split_version = current_version.split("-")
                    if len(split_version) > 1:
                        split_version[-2] += "~rtm"
                    else:
                        split_version[0] += "~rtm"
                    new_version = "-".join(split_version)
                elif (source_archive.distribution.name == "ubuntu-rtm" and
                        dest.distribution.name == "ubuntu" and
                        "~rtm" in current_version):
                    new_version.replace("~rtm", "", 1)

            if packagemanager.is_upstream_version_for_series_in_dest(
                    pkg,
                    new_version,
                    series,
                    src_ppa):
                logging.info(
                    "Package {} seems to be up-to-date in the silo PPA "
                    "(version {})".format(pkg, new_version))
                continue

            dest_source_package = packagemanager.get_source_package_from_dest(
                pkg, source_archive, current_version, source_series, 'build')
            if dest_source_package:
                dest_source_package = os.path.abspath(dest_source_package)
            os.chdir(dest_source_package)

            # here's a tricky hack... in some cases we might need to re-upload
            # existing packages that got deleted since LP does not allow
            # same-version re-uploads, we need to actually bump the ubuntu
            # version number for it
            version_in_ppa = packagemanager.get_current_version_for_series(
                pkg, series.name, dest=src_ppa, all_packages=True)
            if version_in_ppa == new_version:
                new_version = packagemanager.increment_ubuntu_package_version(
                    new_version)

            # modify the changelog
            changelog = open("debian/changelog", "r").readlines()
            changelog_split = changelog[0].split(" ")
            changelog_split[1] = "({})".format(new_version)
            changelog_split[2] = "{};".format(series.name)
            changelog[0] = " ".join(changelog_split)
            with open("debian/changelog", "w") as f:
                for line in changelog:
                    f.write(line)

            # rename the tarball, if needed
            if new_version != current_version:
                os.chdir("..")
                for file_name in os.listdir('.'):
                    if pkg in file_name and ".tar." in file_name:
                        orig_tarball = file_name
                new_upstream_version = packagemanager.get_upstream_version(
                    new_version)
                current_upstream_version = packagemanager.get_upstream_version(
                    current_version)
                new_orig_tarball = orig_tarball.replace(
                    current_upstream_version, new_upstream_version)
                os.rename(orig_tarball, new_orig_tarball)

                os.chdir(dest_source_package)

            # to use our building tools, we actually need to hack around and
            # make a bzr repository (for bzr bd -S)
            logging.debug("Creating the bzr repository")
            subprocess.call(["bzr", "init", "."])
            subprocess.call(["bzr", "add", "."])

            # build the source package
            logging.info(
                "Preparing source package for the sync of {} version "
                "{}".format(pkg, new_version))
            packagemanager.build_source_package(series.name, new_version)

            # save infos for uploading later on.
            src_pkgs_to_upload[pkg] = (
                new_version, "", -1, "", "{}/..".format(dest_source_package))

            # go back to the main directory
            os.chdir(main_dir)

    # do the actual upload to the ppas
    logging.info("Uploading components to the ppa")
    for source_package_name in src_pkgs_to_upload:
        (version,
         branch,
         dest_tip_bzr_rev,
         prev_version,
         path_to_upload_dir) = src_pkgs_to_upload[source_package_name]
        packagemanager.upload_package(
            source_package_name,
            version,
            src_ppa_name,
            path_to_upload_dir)
        # Save infos for other jobs, in case it's a locally built package
        if branch:
            tools.save_project_config(
                source_package_name,
                branch,
                dest_tip_bzr_rev,
                prev_version,
                version)

    # look at the ppa and generate .projects
    logging.info(
        "Looking at the ppa {} ({}) content".format(
            src_ppa.web_link, series.name))
    # as we recreate all .projects file for those direct uploads, let's clean
    # their metadata
    for source in silo_config["sources"]:
        clean_source(source)
    direct_upload_not_in_list = []
    try:
        sources_list = []
        for status in ("Published", "Pending"):
            sources_list.extend(
                list(src_ppa.getPublishedSources(
                    distro_series=series,
                    status=status)))

        for pkg in sources_list:
            if pkg.source_package_name in silo_config["mps"]:
                continue
            if pkg.source_package_name in silo_config["sources"]:
                long_func = packagemanager.get_current_version_for_series
                dest_current_version = long_func(
                    pkg.source_package_name, series.name, dest=dest)
                # if a direct source pacakge upload was already published,
                # ignore it

                published_sources = tools.get_published_to_distro_projects()
                try:
                    if dest_current_version in published_sources[
                            pkg.source_package_name]:
                        continue
                except KeyError:
                    # scan the unpublished upload
                    tools.save_project_config(
                        pkg.source_package_name,
                        "",
                        "",
                        dest_current_version,
                        pkg.source_package_version)
                    # download as well the .dsc file for watch-ppa
                    source_dir = packagemanager.get_source_package_from_dest(
                        pkg.source_package_name,
                        src_ppa,
                        pkg.source_package_version,
                        series.name)
                    shutil.copy(
                        glob.glob(
                            os.path.join(
                                os.path.dirname(source_dir), '*.dsc')
                        )[0],
                        os.path.abspath('.'))
            else:
                direct_upload_not_in_list.append(pkg.source_package_name)
    except lazr.restfulclient.errors.BadRequest:
        pass  # no package in the ppa yet

    if direct_upload_not_in_list:
        message = ("Some packages ({}) that are not in the MP and not in "
                   "sources list have been detected in the ppa. Please remove "
                   "them from the ppa or ask for a reconfiguration.".format(
                       ", ".join(direct_upload_not_in_list)))
        logging.error(message)
        silomanager.set_config_status(
            silo_config,
            silomanager.SILO_STATE_BUILD_FAILED,
            "Build failed: " + message)
        return 1

    # watch ppa
    silomanager.set_config_status(
        silo_config, silomanager.SILO_STATE_BUILDING, "Building", ping=False)
    logging.info("Check ppa status in {}".format(src_ppa.web_link))
    dest_archive = launchpadmanager.get_resource_from_token(
        silo_config["global"]["dest"])
    env.DISTRIBUTION = dest_archive.distribution.name
    env.SERIES = series.name
    env.PPA = "{}/{}/{}".format(
        src_ppa.owner.name,
        dest_archive.distribution.name,
        src_ppa.name)
    if not launchpadmanager.is_dest_distro_archive(
            silo_config["global"]["dest"]):
        env.DEST = dest_ppa_name
    if watch_ppa.main() != 0:
        silomanager.set_config_status(
            silo_config,
            silomanager.SILO_STATE_BUILD_FAILED,
            "Build failed: Some packages failed to build.")
        return 1
    return 0

if __name__ == '__main__':
    if '--help' in sys.argv or '-h' in sys.argv:
        print(__doc__)
    else:
        sys.exit(main())
