#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (C) 2014 Canonical
#
# Authors:
#  Didier Roche
#  Łukasz 'sil2100' Zemczak
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""CI Train Merge & Clean Script

Merge all branches to merge, clean the ppa and free the silo slot.

Environment variables:

ONLY_FREE_SILO
    Only free the ppa and don't merge/check anything.

IGNORE_MISSINGPROJECTS
    Ignore if some projects that are going to be merged back were not
    published. No merge will then be processed for those.

IGNORE_PACKAGES_NOTINDEST
    Ignore if some packages are not published in the destination.

IGNORE_STEP
    Ignore if previous step didn't successfully end.

DEBUG
    Enable debug infos
"""

import logging
import os
import re
import shutil
import sys
import time
import urllib2
import ConfigParser

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from cupstream2distro import (
    branchhandling,
    launchpadmanager,
    packageinppamanager,
    packagemanager,
    silomanager,
)
from cupstream2distro.settings import (
    PROPOSED_URL,
    SILO_PUBLISHED,
    SILO_DONE,
    BRANCH_URL,
    PROJECT_CONFIG_SUFFIX,
    TIME_BETWEEN_PPA_CHECKS,
)
from cupstream2distro.utils import ignored, env


def merge_and_reconciliate(silo_config, packages_in_dest):
    """Merge and eventually reconcialiate all branches"""
    logging.info(
        "Check the revision from all merged branch in trunk didn't "
        "change. If changed, try to reconciliate them back.")
    reconciliate_all_branches = True
    base_path = os.path.abspath('.')
    series = launchpadmanager.get_resource_from_token(
        silo_config["global"]["series"]).name
    dest = launchpadmanager.get_resource_from_token(
        silo_config["global"]["dest"])

    for source in silo_config["mps"]:
        source_branch_uri = os.path.join(base_path, source)
        target_branch = launchpadmanager.get_resource_from_token(
            silo_config['mps'][source][0]).target_branch

        version = packages_in_dest[source]
        config = ConfigParser.RawConfigParser()
        config.read("{}.{}_{}".format(source, PROJECT_CONFIG_SUFFIX, version))
        stored_tip_rev = config.get('Branch', 'rev')
        tip_rev = str(target_branch.revision_count)

        # prepare and remerge
        if tip_rev != stored_tip_rev:
            logging.info(
                "New revision ({}) in {} when prepared is based on {}. Trying "
                "to merge it for you.".format(tip_rev, source, stored_tip_rev))
            if not branchhandling.merge_branch(
                    source_branch_uri,
                    target_branch.web_link,
                    "Resync trunk"):
                proposed_branch = PROPOSED_URL.format(
                    target_branch.project.name,
                    dest.distribution.name, series)
                logging.error(
                    "We can't reconciliate your proposed branch (pushed to "
                    "{}) with the new commits in target (merge "
                    "conflicts).".format(proposed_branch))
                reconciliate_all_branches = False
    if not reconciliate_all_branches:
        message = (
            "Some branches couldn't be reconciliated with the changes "
            "happening meaning in destination branch. Please revert your "
            "changes in trunk and relaunch this job.")
        logging.error(message)
        silomanager.set_config_status(
            silo_config,
            silomanager.SILO_STATE_MERGE_FAILED,
            "Can't merge: " + message)
        return 1

    logging.info("Pushing the branches to the parent ones")
    push_all_branches = True
    for source in silo_config["mps"]:
        branch_uri = os.path.join(base_path, source)
        target_branch_link = launchpadmanager.get_resource_from_token(
            silo_config['mps'][source][0]).target_branch.web_link
        logging.info("Trying to push to {}".format(target_branch_link))
        if not branchhandling.push_to_branch(branch_uri, target_branch_link):
            url = BRANCH_URL.format(
                source,
                "{}-recup".format(dest.distribution.name))
            logging.error(
                "Can't push this branch to {}, pushing temporarly "
                "to {}.".format(target_branch_link, url))
            branchhandling.push_to_branch(branch_uri, url, overwrite=True)
            push_all_branches = False
    if not push_all_branches:
        message = (
            "Some branches couldn't be pushed to the destination. Please "
            "pull them from the url link above and relaunch this job with "
            "only free slot flag.")
        logging.error(message)
        silomanager.set_config_status(
            silo_config,
            silomanager.SILO_STATE_MERGE_FAILED,
            "Can't merge: " + message)
        return 1


def main():
    logging.basicConfig(
        level=logging.DEBUG if env.DEBUG == 'true' else logging.INFO,
        format="%(asctime)s %(levelname)s %(message)s")

    logging.debug(str(env))

    silo_config = silomanager.load_config()
    if (silo_config["global"]["status"]["state"] ==
            silomanager.SILO_STATE_MERGING):
        logging.error(
            "The merge and clean job seems to have been already ran. "
            "Skipping.")
        return 1
    silomanager.set_config_status(
        silo_config,
        silomanager.SILO_STATE_MERGING,
        "Merging",
        ping=False)

    if env.ONLY_FREE_SILO != 'true':
        if silomanager.get_config_step(silo_config) != SILO_PUBLISHED:
            if env.IGNORE_STEP == 'true':
                logging.info(
                    "Last step didn't finish successfully, but the "
                    "ignore flag is set.")
            else:
                logging.error(
                    "Last step didn't finish successfully. You need to "
                    "either ignore that the previous step didn't finished "
                    "successfully or ensuring that prepare, build and "
                    "checked passed.")
                silomanager.set_config_status(
                    silo_config,
                    silomanager.SILO_STATE_MERGE_FAILED,
                    "Can't merge: wrong status or parameters for job.")
                return 1
            # reset it
            if not silomanager.set_config_step(silo_config, SILO_PUBLISHED):
                logging.error(
                    "Couldn't reset the step in the configuration file.")
                silomanager.set_config_status(
                    silo_config,
                    silomanager.SILO_STATE_MERGE_FAILED,
                    "Merge failed: couldn't reset status or parameters "
                    "for job.")
                return 1

    # get latest uploaded version for all packages
    logging.info(
        "Getting all sources that were uploaded with their latest version")
    packages_in_dest = {}
    packages = packageinppamanager.get_packages_and_versions_uploaded()
    for (source, version) in packages:
        previous_scanned_uploaded_version = packages_in_dest.get(source, "0")
        if packagemanager.is_version1_higher_than_version2(
                version,
                previous_scanned_uploaded_version):
            packages_in_dest[source] = version

    # check that all packages have reached their destination
    result = packagemanager.check_package_reached_destination(
        silo_config,
        packages_in_dest,
        env.IGNORE_MISSINGPROJECTS == 'true',
        env.IGNORE_PACKAGES_NOTINDEST == 'true',
        for_merge_and_clean=True)
    (all_pkg_in_dest, message) = result
    if not all_pkg_in_dest and env.ONLY_FREE_SILO != 'true':
        silomanager.set_config_status(
            silo_config,
            silomanager.SILO_STATE_MIGRATING,
            message)
        return 1

    if env.ONLY_FREE_SILO != 'true':
        merge_and_reconciliate(silo_config, packages_in_dest)
    else:
        logging.warning(
            "Only free silo was requested, not merging potential branches")

    # set the silo as done
    silomanager.set_config_step(silo_config, SILO_DONE)

    if env.ONLY_FREE_SILO != 'true':
        silomanager.set_config_status(
            silo_config,
            silomanager.SILO_STATE_CLEANING,
            "Landed. Cleaning silo.")
    # we gave up
    else:
        if not all_pkg_in_dest:
            silomanager.set_config_status(
                silo_config,
                silomanager.SILO_STATE_CLEANING,
                "Gave up this landing. Cleaning silo.")
        else:
            silomanager.set_config_status(
                silo_config,
                silomanager.SILO_STATE_CLEANING,
                "Landed (without branch merged). Cleaning silo.")

    ppa = launchpadmanager.get_resource_from_token(
        silo_config["global"]["ppa"])
    series = launchpadmanager.get_resource_from_token(
        silo_config["global"]["series"])
    ppa_base_path = "http://ppa.launchpad.net/{}/{}/{}/dists/{}/main/".format(
        ppa.owner.name, ppa.name, ppa.distribution.name, series.name)
    logging.info(
        "Clean now the source ppa {} in {}".format(ppa.web_link, series.name))

    for pkg in ppa.getPublishedSources(distro_series=series):
        if pkg.status != "Deleted":
            logging.info(
                "Requesting deletion for {}".format(pkg.source_package_name))
            pkg.requestDeletion()

    # Source publication is lying, checking directly the repository
    remaining_packages = False
    try:
        archs_page = urllib2.urlopen(ppa_base_path).read().decode('utf-8')
        remaining_packages = True
    except urllib2.HTTPError:
        # no remaining packages, so we won't have any further check
        archs_page = ""
    binary_pattern = re.compile('href="(binary-.*)/">')
    url_lists = []
    for url in binary_pattern.findall(archs_page):
        url_lists.append("{}{}/Packages".format(ppa_base_path, url))
    url_lists.append("{}source/Sources".format(ppa_base_path))
    while(remaining_packages):
        time.sleep(TIME_BETWEEN_PPA_CHECKS)
        logging.info("Checking if ppa is empty.")
        remaining_packages = False

        for url in url_lists:
            if len(urllib2.urlopen(url).read().decode('utf-8')) != 0:
                logging.info(
                    "{} still have some packages published".format(url))
                remaining_packages = True

    # Finally, nuke the silo dir so we don't waste space on the server
    current_dir = os.path.abspath(".")
    with ignored(OSError):
        shutil.rmtree(current_dir)
        os.mkdir(current_dir)
    silomanager.remove_status_file(silo_config["siloname"])
    return 0

if __name__ == '__main__':
    if '--help' in sys.argv or '-h' in sys.argv:
        print(__doc__)
    else:
        sys.exit(main())
