#!/usr/bin/python -u
# -*- coding: utf-8 -*-
# Copyright (C) 2014 Canonical
#
# Authors:
#  Didier Roche
#  Łukasz 'sil2100' Zemczak
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""CI Train Prepare Silo Job

Prepare and assign a free silo from a ppa.
You must specify either MPs or sources.

Environment variables:

DEBUG
    Enable debug infos

DEST
    Use this destination PPA instead of {series}-proposed

DISTRIBUTION
    Which distribution to target (default: ubuntu)

IGNORECONFLICTS
    Allow same source package in multiple silos. You must coordinate both silos
    to ensure that when one publishes, the other one rebuilds, otherwise the
    second publication will revert the work done in the initial publication.

LANDERS
    Landers who are owners of that request (should be IRC nicks for pinging)

MERGE_PROPOSALS
    Merge proposal urls (space separated)

RECONFIGURE_SILO
    Reconfigure an existing silo

REQUEST_ID
    Update/Configure this request id

RESTRICTED
    Only add MPs for already-locked source packages (won't allow new sources)

SERIES
    Series used to build the package

SOURCES
    Source packages for manual upload (space separated)

SYNC_REQUEST
    Copy packages from a different silo/archive rather than building MPs.

USE_PREPROD_SILO
    Assign a silo with preproduction CI Train code
"""

import lazr
import logging
import os
import sys

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from cupstream2distro import (
    branchhandling,
    launchpadmanager,
    packageinppamanager,
    silomanager,
)
from cupstream2distro.settings import (
    SILO_BUILDPPA_SCHEME,
    SILOS_DIR,
    SILO_NAME_LIST,
    SILO_EMPTY,
)
from cupstream2distro.tools import clean_source
from cupstream2distro.utils import ignored, env


def check_not_adding_new_components(project_name,
                                    allowed_components_from_previous_config,
                                    config,
                                    silo_path):
    """Check that a self reconfigure doesn't add new components"""

    if project_name not in allowed_components_from_previous_config:
        message = ("{} was not in the initial list of components for that "
                   "silo. You can't reconfigure the silo yourself. "
                   "Please ask the landing team to reconfigure it for "
                   "you.".format(project_name))
        logging.error(message)
        silomanager.set_config_step(config, -1, uri=silo_path)
        silomanager.set_config_status(
            config,
            silomanager.SILO_STATE_RECONFIGURE_FAILED,
            "Reconfigure failed: " + message,
            uri=silo_path)
        return True
    return False


def main():
    logging.basicConfig(
        level=logging.DEBUG if env.DEBUG == 'true' else logging.INFO,
        format="%(asctime)s %(levelname)s %(message)s")

    logging.debug(str(env))

    if not env.REQUEST_ID:
        logging.error("Request id is mandatory")
        return 1
    if not (env.MERGE_PROPOSALS or env.SOURCES or env.SYNC_REQUEST):
        logging.error(
            "Must specify either MERGE_PROPOSALS, SOURCES, or SYNC_REQUEST")
        return 1
    if env.RESTRICTED == 'true' and env.RECONFIGURE_SILO != 'true':
        logging.error(
            "Restricted mode should only be used to reconfigure a silo")
        return 1

    project_urls = env.MERGE_PROPOSALS.split()
    project_urls.extend(env.SOURCES.split())

    distribution = env.DISTRIBUTION or 'ubuntu'
    # make sure the distribution parameter is one of the two supported right
    # now
    if distribution not in SILO_NAME_LIST:
        logging.error(
            "Distribution {} is currently not supported in the CI "
            "Train. Not assigning".format(distribution))
        return 1

    # create all silo names
    with ignored(OSError):
        os.mkdir(SILOS_DIR)
    for silo_name in SILO_NAME_LIST[distribution]:
        with ignored(OSError):
            os.mkdir(os.path.join(SILOS_DIR, silo_name))

    # first, check that a silo is available
    dest = None
    series = None

    existing_config = silomanager.find_silo_config_for_request_id(
        env.REQUEST_ID, SILOS_DIR)

    # check provided input matches
    if env.RECONFIGURE_SILO != 'true' and existing_config:
        logging.error(
            "You didn't ask to reconfigure a silo, but we found it "
            "assigned to {}. Please ensure to check the missing "
            "option if this is really what you want to "
            "do.".format(existing_config['siloname']))
        return 1
    if env.RECONFIGURE_SILO == 'true' and not existing_config:
        logging.error(
            "You asked to reconfigure request id {}. But we couldn't "
            "find it in any existing silo for it. Are you sure you "
            "don't want rather to assign a new silo with "
            "it?".format(env.REQUEST_ID))
        return 1

    if env.RECONFIGURE_SILO == 'true' and existing_config:
        silo_name = existing_config['siloname']
        logging.info(
            "Reconfiguring {} as an existing silo for request "
            "{}".format(silo_name, env.REQUEST_ID))
        # set the same destination, series
        dest = launchpadmanager.get_resource_from_token(
            existing_config["global"]["dest"])
        series = launchpadmanager.get_resource_from_token(
            existing_config["global"]["series"])
        # ensure that the series matches the old series field or yell
        if env.SERIES and series.name != env.SERIES:
            logging.error(
                "You provided a series in the reconfigure {} which is "
                "different from the original series that request was for "
                "{}. To ensure this is what's wanted, it's preferable to "
                "free this silo and request a new one for the new "
                "series.".format(env.SERIES, series))
            return 1
        allowed_components_from_previous_config = existing_config[
            'mps'].keys() + existing_config['sources']
        logging.debug("Components in previous configure: " +
                      ", ".join(allowed_components_from_previous_config))
    else:
        logging.info("Checking that a silo is available")
        silo_name = silomanager.return_first_available_silo(
            SILOS_DIR,
            preprod=env.USE_PREPROD_SILO == 'true',
            distribution=env.DISTRIBUTION)
        if not silo_name:
            logging.error("No silo is available")
            return 1

    # silo attribution (made automatically or explicitely)
    if not series:
        if not env.SERIES:
            logging.error(
                "You need to provide a series to attribute a new silo")
            return 1
        series = launchpadmanager.get_series(env.SERIES, env.DISTRIBUTION)
        if env.DEST:
            dest = launchpadmanager.get_ppa(env.DEST)
        else:
            dest = launchpadmanager.get_distribution_archive(env.DISTRIBUTION)

    logging.info("Configuring for {} ({}) in {}".format(
        ", ".join(project_urls), series.name, dest.name))
    config = {"requestid": env.REQUEST_ID,
              "global": {"dest": dest.self_link,
                         "series": series.self_link,
                         "step": "",
                         "status": ""},
              "landers": env.LANDERS.split(),
              "siloname": silo_name,
              "sources": [],
              "mps": {}
              }

    if env.SYNC_REQUEST:
        split = env.SYNC_REQUEST.split(",")
        # check if it's a sync from a PPA or a distro
        if split[0].startswith("ppa:"):
            logging.info("Sync from a PPA ({}) requested".format(split[0]))

            ppa_name = split[0].split(":")[1]
            if ppa_name[0] == "~":
                ppa_name = ppa_name[1:]
            source_archive = launchpadmanager.get_ppa(ppa_name)
            if len(split) > 1:
                source_series = split[1]
        elif split[0].isdigit():
            # for silo synces we only handle ubuntu and ubuntu-rtm
            source_distro = (
                "ubuntu" if env.DISTRIBUTION == "ubuntu-rtm" else
                "ubuntu-rtm")
            source_landing = "landing-{:03d}".format(int(split[0]))
            source_archive = launchpadmanager.get_ppa(
                "ci-train-ppa-service/{}/{}".format(
                    source_distro,
                    source_landing))
            source_series = (
                "utopic" if env.DISTRIBUTION == "ubuntu-rtm" else
                "14.09")

            logging.info(
                "Sync from a silo ({}/{}) requested".format(
                    source_distro,
                    source_landing))

            # read the configuration of the source silo
            source_config = silomanager.load_config(
                os.path.join(SILOS_DIR, source_distro, source_landing))
            if not source_config:
                logging.error(
                    "The source silo is either not accessible or not "
                    "assigned, please fill in a different sync request")
                return 1

            config["source_sync_from_silo"] = source_config["requestid"]
            if not env.SOURCES:
                logging.info(
                    "No sources list provided, syncing whole silo contents")
                config["source_sync_all_contents"] = True
        else:
            logging.info(
                "Sync from a distribution archive ({}) requested".format(
                    split[0]))

            source_archive = launchpadmanager.get_distribution_archive(
                split[0])
            if len(split) > 1:
                source_series = split[1]

        if not source_archive:
            logging.error(
                "Sync silo requested but source archive is not valid "
                "({})".format(split[0]))
            return 1

        if not source_series:
            logging.error(
                "Sync silo requested but no series given. Request was the "
                "following: {}".format(env.SYNC_REQUEST))
            return 1

        if not env.SOURCES and "source_sync_all_contents" not in config:
            logging.error(
                "Sync silo requested and no sources given, "
                "which is not allowed for non-silo synces")
            return 1

        source_archive_link = source_archive.self_link

        config["source_archive"] = source_archive_link
        config["source_series"] = source_series

    silo_path = os.path.join(SILOS_DIR, silo_name)
    logging.info("Silo path: {}, silo name: {}".format(silo_path, silo_name))

    if env.SOURCES:
        logging.info(
            "Additional packages will be directly uploaded to the ppa: "
            "{}".format(", ".join(env.SOURCES.split())))
        config["sources"] = env.SOURCES.split()

    one_project_already_in_use = False
    source_packages_seen = set()
    # grab the project name of all MP
    for mp_url in env.MERGE_PROPOSALS.split():
        logging.debug('Checking MP: ' + mp_url)
        mp = launchpadmanager.get_resource_from_url(mp_url)
        try:
            long_func = branchhandling.get_source_package_name_from_branch
            source_package_name = long_func(mp.target_branch.web_link)
        except AttributeError:
            logging.error(
                "{} doesn't seem to be a valid merge proposal "
                "url".format(mp_url))
            return 1
        project_name = mp.target_branch.project.name
        logging.debug('Found project name ' + project_name)

        # ensure a self reconfiguring doesn't add new projects
        if env.RECONFIGURE_SILO == env.RESTRICTED == 'true':
            if check_not_adding_new_components(
                    source_package_name,
                    allowed_components_from_previous_config,
                    existing_config,
                    silo_path):
                return 1

        if (source_package_name != project_name):
            message = ("Project name ({}) doesn't align with the source "
                       "package name ({})".format(project_name,
                                                  source_package_name))
            logging.warning(message)

        config["mps"].setdefault(source_package_name, []).append(mp.self_link)

        logging.debug('Seen: ' + str(source_packages_seen))
        if (source_package_name not in source_packages_seen):
            source_packages_seen.add(source_package_name)
            # check that any of those components are part of any current
            # transaction
            if not silomanager.is_project_not_in_any_configs(
                    mp.target_branch.project.name,
                    series,
                    dest,
                    SILOS_DIR,
                    silo_name,
                    dont_error_but_warn=env.IGNORECONFLICTS == 'true'):
                one_project_already_in_use = True

    # check that additional sources are not as well already in use
    for source in config["sources"]:
        logging.debug('Checking source: ' + source)

        # FIXME: This looks like some copypasta from above, refactor it pls.
        # ensure a self reconfiguring doesn't add new projects
        if env.RECONFIGURE_SILO == env.RESTRICTED == 'true':
            if check_not_adding_new_components(
                    source,
                    allowed_components_from_previous_config,
                    existing_config,
                    silo_path):
                return 1

        logging.debug('Seen: ' + str(source_packages_seen))
        if (source not in source_packages_seen):
            source_packages_seen.add(source)
            if not silomanager.is_project_not_in_any_configs(
                    source,
                    series,
                    dest,
                    SILOS_DIR,
                    silo_name,
                    dont_error_but_warn=env.IGNORECONFLICTS == 'true'):
                one_project_already_in_use = True

    if one_project_already_in_use:
        message = ("One or more projects are already in use for the same "
                   "destination and series in another silo (see above).")
        if env.IGNORECONFLICTS == 'true':
            logging.warning(
                message + " However, the ignore flag was set. Please inform "
                "both silos owners directly.")
        else:
            logging.error(message)
            return 1

    # check that all MPs for a source are targetting the same branch
    logging.info(
        "Checking that all target branch for the same source in mps are "
        "targetting the same branches")
    all_same_dest = True
    for source in config["mps"]:
        first_mp = None
        for mp_link in config['mps'][source]:
            mp = launchpadmanager.get_resource_from_token(mp_link)
            if not first_mp:
                first_mp = mp
            else:
                current_branch_link = mp.target_branch.web_link
                if current_branch_link != first_mp.target_branch.web_link:
                    logging.error(
                        "{} is targetting {} while {} is targetting "
                        "{}".format(
                            mp.web_link,
                            current_branch_link,
                            first_mp.web_link,
                            first_mp.target_branch.web_link))
                    all_same_dest = False

    if not all_same_dest:
        logging.error(
            "Not all MP targetting the same source package are targetting "
            "the same branch. Please recheck your configuration.")
        return 1

    ppa_uri = SILO_BUILDPPA_SCHEME.format(env.DISTRIBUTION, silo_name.split(
        '/')[-1])  # make sure the silo name does not include the distro prefix
    try:
        ppa = launchpadmanager.get_ppa(ppa_uri)
    except lazr.restfulclient.errors.NotFound:
        logging.error("PPA {} not found".format(ppa_uri))
        return 1
    config["global"]["ppa"] = ppa.self_link
    if not silomanager.save_config(config, silo_path):
        return 1
    logging.info(
        "Assigning \"{}\" as silo. Build PPA is available at "
        "{}".format(silo_name, ppa.web_link))

    # if reconfigure, ensure we remove all projects file and packages from the
    # ppa that are not to take into account anymore
    if env.RECONFIGURE_SILO == 'true':
        os.chdir(silo_path)
        logging.info("Check and clean all removed sources")
        all_silo_projects = silomanager.get_all_projects(config)
        logging.debug("all silo projects {}".format(all_silo_projects))
        # remove all files related to sources that are not in the configuration
        # anymore as well as all additional uploads in the ppa
        projects_list_to_eventually_clean = set()

        # first source: ppa
        result = packageinppamanager.get_all_packages_uploaded()
        for (source, version, rev, branch) in result:
            logging.debug("Detected in the ppa: {}".format(source))
            projects_list_to_eventually_clean.add(source)
        # second source: package that were in the previous config
        for source in allowed_components_from_previous_config:
            logging.debug("From previous config: {}".format(source))
            projects_list_to_eventually_clean.add(source)
        for source in projects_list_to_eventually_clean:
            if source not in all_silo_projects:
                logging.info(
                    "Removing {} from ppa and configuration".format(source))
                clean_source(source)
                for pkg in ppa.getPublishedSources(
                        exact_match=True,
                        distro_series=series,
                        source_name=source):
                    pkg.requestDeletion()
        # NOTE: we don't wait for the removal to be published. Shouldn't be an
        # issue

    silomanager.set_config_step(config, SILO_EMPTY, uri=silo_path)
    silomanager.set_config_status(
        config,
        silomanager.SILO_STATE_NONE,
        "Silo ready to build packages",
        uri=silo_path,
        add_url=False)
    return 0

if __name__ == '__main__':
    if '--help' in sys.argv or '-h' in sys.argv:
        print(__doc__)
    else:
        sys.exit(main())
