#!/usr/bin/python -u
# -*- coding: utf-8 -*-
# Copyright (C) 2014 Canonical
#
# Authors:
#  Didier Roche
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""CI Train Prepare Silo Using Spreadsheet Info Job

Prepare and assign a free silo from a ppa.

Environment variables:

DEBUG
    Enable debug infos

DEST
    Use this destination PPA instead of {series}-proposed

DISTRIBUTION
    Which distribution to target (default: ubuntu)

IGNORECONFLICTS
    Allow same source package in multiple silos. You must coordinate both silos
    to ensure that when one publishes, the other one rebuilds, otherwise the
    second publication will revert the work done in the initial publication.

LANDERS
    Landers who are owners of that request (should be IRC nicks for pinging)

RECONFIGURE_FOR_SILO
    Name of the silo you want to reconfigure

RECONFIGURE_SILO
    Reconfigure an existing silo (true or false)

REQUEST_ID
    Update/Configure this request id

RESTRICTED
    Only add MPs for already-locked source packages (won't allow new sources)

SERIES
    Series used to build the package

SYNC_REQUEST
    Copy packages from a different silo/archive rather than building MPs.

USE_PREPROD_SILO
    Assign a silo with preproduction CI Train code
"""

import csv
import logging
import os
import requests
import sys
import tempfile
import re

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from citrain import prepare_silo
from cupstream2distro.settings import SPREADSHEET_CSV_URL
from cupstream2distro.tools import parse_and_clean_entry
from cupstream2distro.utils import env


class COL:
    pass


def main():
    logging.basicConfig(
        level=logging.DEBUG if env.DEBUG == 'true' else logging.INFO,
        format="%(asctime)s %(levelname)s %(message)s")

    if env.RECONFIGURE_FOR_SILO:
        logging.info("Reconfiguration of silo requested for {}".format(
            env.RECONFIGURE_FOR_SILO))
    else:
        if not env.REQUEST_ID:
            logging.error("Request id is mandatory")
            return 1

    logging.info(
        "Getting data from the spreadsheet for request "
        "{}".format(env.REQUEST_ID))

    # proceed and download the csv version of the spreadsheet (only way to get
    # data anonymously)
    response = requests.get(SPREADSHEET_CSV_URL)
    if response.status_code != 200:
        logging.error(
            "Couldn't fetch {} and so, couldn't proceed your question. Can "
            "be due to an temporary network hiccup, please retry "
            "later.".format(SPREADSHEET_CSV_URL))
        return 1

    # store the answer in a temp file
    (fd, filename) = tempfile.mkstemp()
    with open(filename, 'w') as f:
        f.write(response.content.encode('utf8'))

    request_found = False
    # use to ensure the unicity of silo name (shouldn't happen apart from a bad
    # sync on the spreadsheet)
    unique_silo_name = True
    with open(filename, 'rb') as f:
        content = csv.reader(f)

        # First two rows are misc header data that don't conform to the columns
        content.next() and content.next()

        # Map column header names to column numbers
        for i, heading in enumerate(content.next()):
            setattr(COL, re.split('\W', heading.lower())[0], i)

        # Now iterate over the remaining rows, which are just the actual data
        for row in content:
            if env.RECONFIGURE_FOR_SILO:
                if row[COL.assigned] == env.RECONFIGURE_FOR_SILO:
                    # second matching
                    if request_found:
                        logging.error(
                            "{} is available more than once in the "
                            "spreadsheet. Please request to fix some wrong "
                            "metadata due to a bad automated sync.".format(
                                env.RECONFIGURE_FOR_SILO))
                        unique_silo_name = False
                    env.REQUEST_ID = row[COL.request]
                    raw_mps = row[COL.merge]
                    raw_sources = row[COL.additional]
                    raw_landers = row[COL.lander]
                    distribution = row[COL.target]
                    request_found = True
            else:
                if row[COL.request] == env.REQUEST_ID:
                    raw_mps = row[COL.merge]
                    raw_sources = row[COL.additional]
                    raw_landers = row[COL.lander]
                    distribution = row[COL.target]
                    request_found = True
                    break
    os.remove(filename)

    if not unique_silo_name:
        return 1

    # ensure we found anything
    if not request_found:
        logging.error(
            "We couldn't find any request in the spreadsheet matching your "
            "demand. Ensure with the landing team there is a matching "
            "request in.")
        return 1

    mps = parse_and_clean_entry(raw_mps)
    sources = parse_and_clean_entry(raw_sources, comma_as_sep=False)
    landers = parse_and_clean_entry(raw_landers, slash_as_sep=True)

    if mps:
        env.MERGE_PROPOSALS = ' '.join(mps)
    if sources:
        env.SOURCES = ' '.join(sources)
        if sources[0].startswith("sync:"):
            # in case we have a sync request, remove it from the sources list
            # and mark the silo properly
            env.SYNC_REQUEST = sources[0][5:]
            env.SOURCES = ' '.join(sources[1:])
    if landers:
        env.LANDERS = ' '.join(landers)
    if distribution:
        split = distribution.split('/')
        env.DISTRIBUTION = split[0]
        if len(split) > 1:
            env.SERIES = split[1]

    if prepare_silo.main() != 0:
        logging.error("Couldn't assign or reconfigure silo")
        return 1
    return 0

if __name__ == '__main__':
    if '--help' in sys.argv or '-h' in sys.argv:
        print(__doc__)
    else:
        sys.exit(main())
