#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (C) 2012-2014 Canonical
#
# Authors:
#  Didier Roche
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""CI Train Setup CI Train Job

Create/Update the configuration of the Jenkins jobs for CI Train

Environment variables:

ALL_SILOS
    Configure jenkins jobs associated with all silos

ONLY_SILOS
    Only configure jenkins jobs associated with these silos

PREPARE
    Configure prepare job as well

CHECKPUBLICATIONMIGRATION
    Configure check-publication-migration job as well

DEPLOY_PREPROD_CITRAIN
    Deploy custom branch for preproduction code of CI Train

DEPLOY_PROD_CITRAIN
    Deploy latest trunk code as production CI Train system

DEPLOY_CHROOT
    Deploy the chroot maintenance job(s)

DEPLOY_DEPLOY_CITRAIN
    Deploy the latest version of this job as well (yo dawg, I heard you like to
    deploy, so we deployed your deploy script so you can deploy your deploy
    script while you deploy)

DEBUG
    Enable debug infos
"""

import jenkins
import os
import logging
import yaml
import subprocess
import sys
import urllib2
import shutil

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from cupstream2distro.utils import env
from cupstream2distro.settings import (
    CITRAIN_BINDIR,
    CITRAIN_PREPROD_BINDIR,
    SILO_NAME_LIST,
    SILO_PREPROD_NAME_LIST,
    SILOS_DIR,
    SILO_RSYNCDIR,
    CITRAIN_SETTINGS,
)

ALL_SILOS = SILO_NAME_LIST['ubuntu'] + SILO_NAME_LIST['ubuntu-rtm']
ALL_PREPROD_SILOS = (SILO_PREPROD_NAME_LIST['ubuntu'] +
                     SILO_PREPROD_NAME_LIST['ubuntu-rtm'])
DEFAULT_CREDENTIALS = os.path.expanduser('~/.cu2d.cred')
JENKINS_DAILY_CONFIG_NAME = 'citrain'
TEMPLATE_DIR = os.path.join(os.path.dirname(__file__), 'jenkins-templates')


def template(template_name):
    return template_name + '.xml.tmpl'


def load_template(template_name, context):
    with open(os.path.join(TEMPLATE_DIR, template_name)) as templ:
        return templ.read().format(**context)


class JenkinsInstance:
    """Our connection to Jenkins."""

    def __init__(self):
        credentials = None
        credentials = self.load_jenkins_credentials(
            os.path.expanduser(DEFAULT_CREDENTIALS), JENKINS_DAILY_CONFIG_NAME)
        if not credentials:
            logging.error("Couldn't find jenkins credentials.")
            return 1
        self.jenkins_handle = self.get_jenkins_handle(credentials)
        if not self.jenkins_handle:
            logging.error("Could not acquire connection to jenkins.")
            return 1

    def load_jenkins_credentials(self, path, jenkins_name):
        """Load credentials from credentials configuration file"""
        if not os.path.exists(path):
            logging.error("No credential file found in {}".format(path))
            return False

        logging.debug("Loading credentials from {}".format(path))
        cred = yaml.safe_load(file(path, 'r'))
        return False if jenkins_name not in cred else cred[jenkins_name]

    def get_jenkins_handle(self, jenkins_config):
        """Get jenkins object"""
        logging.debug('jenkings_config[url]: ' + jenkins_config['url'])
        if not jenkins_config['url']:
            logging.error("Please provide a URL to the jenkins instance.")
            return False
        if 'username' in jenkins_config:
            jenkins_handle = jenkins.Jenkins(
                jenkins_config['url'],
                username=jenkins_config['username'],
                password=jenkins_config['password'])
        else:
            jenkins_handle = jenkins.Jenkins(jenkins_config['url'])
        return jenkins_handle

    def setup_job(self, jobname, template_name, context={}):
        """ Generate template and create or update jenkins job

        :param jobname: jenkins' job name
        :param template_name: template name
        :param context: additional context
        """
        logging.debug("Generating job: {}".format(jobname))
        context['BINDIR'] = CITRAIN_BINDIR
        for preprod_silo in ALL_PREPROD_SILOS:
            if preprod_silo in jobname:
                context['BINDIR'] = CITRAIN_PREPROD_BINDIR
                break
        context['RSYNC_OUTDIR'] = SILO_RSYNCDIR
        jkcfg = load_template(template_name, context)
        jobname = urllib2.quote(jobname)
        logging.debug('jkcfg: ' + jkcfg)
        if not self.jenkins_handle.job_exists(jobname):
            logging.info("Creating Jenkins Job {}".format(jobname))
            self.jenkins_handle.create_job(jobname, jkcfg)
        else:
            logging.info("Reconfiguring Jenkins Job {}".format(jobname))
            self.jenkins_handle.reconfig_job(jobname, jkcfg)

    def deploy_citrain(self, preprod_branch=None):
        """Deploy latest citrain code or use new code for preprod"""
        cur_dir = os.path.abspath('.')

        if preprod_branch:
            # Deploy preproduction
            preprod_dir = os.path.join(cur_dir, 'citrain-preprod')
            preprod_dir_new = preprod_dir + "_new"
            preprod_dir_old = preprod_dir + "_old"

            # Clean up on start in case last upgrade didn't happen completely
            if os.path.exists(preprod_dir_new):
                shutil.rmtree(preprod_dir_new)
            if os.path.exists(preprod_dir_old):
                shutil.rmtree(preprod_dir_old)

            subprocess.call(['bzr', 'branch', preprod_branch, preprod_dir_new])
            if os.path.exists(preprod_dir):
                os.rename(preprod_dir, preprod_dir_old)
            os.rename(preprod_dir_new, preprod_dir)
        else:
            # Deploy production
            citrain_dir = os.path.join(cur_dir, 'citrain')
            source_dir = 'lp:cupstream2distro'
            if not os.path.isdir(citrain_dir):
                subprocess.call(['bzr', 'branch', source_dir, citrain_dir])
            else:
                os.chdir(citrain_dir)
                subprocess.call(['bzr', 'pull', '--overwrite'])
                os.chdir(cur_dir)

    def setup_silo(self, silo_name):
        """Setup a silo jenkins job"""
        logging.info("Configure {}".format(silo_name))
        context = {
            "SILOS_DIR": SILOS_DIR,
            "SILO_NAME": silo_name,
            "MERGE_AND_CLEAN_TOKEN": ""
        }

        settings = yaml.load(file(CITRAIN_SETTINGS))
        if "auto_merge_token" in settings:
            context["MERGE_AND_CLEAN_TOKEN"] = settings["auto_merge_token"]

        # change the names from 'distro/landing-num' to 'distro-landing-num'
        silo_name = silo_name.replace("/", "-")

        self.setup_job(
            "{}-0-reconfigure".format(silo_name),
            template("reconfigure-silo"),
            context)
        self.setup_job(
            "{}-1-build".format(silo_name),
            template("build"),
            context)
        self.setup_job(
            "{}-2-publish".format(silo_name),
            template("publish"),
            context)
        self.setup_job(
            "{}-3-merge-clean".format(silo_name),
            template("merge-clean"),
            context)
        self.setup_job(
            "{}-4-auto-merge-clean".format(silo_name),
            template("merge-clean"),
            context)


def main():
    logging.basicConfig(
        level=logging.DEBUG if env.DEBUG == 'true' else logging.INFO,
        format="%(asctime)s %(levelname)s %(message)s")

    logging.debug(str(env))

    jenkinstance = JenkinsInstance()

    if env.DEPLOY_PREPROD_CITRAIN:
        logging.info(
            "Deploy latest citrain from branch {} to preprod".format(
                env.DEPLOY_PREPROD_CITRAIN))
        jenkinstance.deploy_citrain(env.DEPLOY_PREPROD_CITRAIN)
    if env.DEPLOY_PROD_CITRAIN == 'true':
        logging.info("Deploy latest citrain code to production")
        jenkinstance.deploy_citrain()
    if env.PREPARE == 'true':
        logging.info("Configure prepare silos job")
        jenkinstance.setup_job(
            "prepare-silo", template("prepare-silo"))
        jenkinstance.setup_job(
            "prepare-silo-manual", template("prepare-silo-manual"))
    if env.CHECKPUBLICATIONMIGRATION == 'true':
        logging.info("Configure check-publication-migration job")
        jenkinstance.setup_job(
            "check-publication-migration",
            template("check-publication-migration"))
    if env.ONLY_SILOS:
        logging.info("Configure a set of silos")
        for silo_name in env.ONLY_SILOS.split():
            if silo_name not in ALL_SILOS:
                logging.error("silo {} doesn't exist in {}".format(
                    silo_name, ALL_SILOS))
                return 1
            jenkinstance.setup_silo(silo_name)
    elif env.ALL_SILOS == 'true':
        logging.info("Configure all silos")
        for silo_name in ALL_SILOS:
            jenkinstance.setup_silo(silo_name)
    if env.DEPLOY_DEPLOY_CITRAIN == 'true':
        logging.info("Configure deploy citrain job")
        jenkinstance.setup_job("deploy-citrain", template("deploy-citrain"))
    if env.DEPLOY_CHROOT == 'true':
        logging.info("Configure chroot maintenance job(s)")
        jenkinstance.setup_job("upgrade-chroot", template("upgrade-chroot"))
    return 0

if __name__ == '__main__':
    if '--help' in sys.argv or '-h' in sys.argv:
        print(__doc__)
    else:
        sys.exit(main())
