#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (C) 2012 Canonical
#
# Authors:
#  Didier Roche
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""CI Train Watch PPA Script

Watch for published package in a ppa

Environment variables:

DISTRIBUTION
    Distribution used to build the package (default: ubuntu)

SERIES
    Series used to build the package

PPA
    PPA to publish this package to (for instance: 'ubuntu-unity/daily-build')

ARCH
    Only consider the provided target

DEST
    Consider this destppa instead of {series}-proposed
"""

import logging
import os
import sys
import time

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from cupstream2distro.utils import env
from cupstream2distro import launchpadmanager, packageinppamanager
from cupstream2distro.packageinppa import PackageInPPA
from cupstream2distro.packagemanager import list_packages_info_in_str
from cupstream2distro.settings import (
    TIME_BETWEEN_PPA_CHECKS,
    TIME_BEFORE_STOP_LOOKING_FOR_SOURCE_PUBLISH,
)


def main():
    logging.basicConfig(level=logging.INFO,
                        format="%(asctime)s %(levelname)s %(message)s")

    distribution = env.DISTRIBUTION or 'ubuntu'
    series = env.SERIES
    ppa = env.PPA
    instance_info = "watching: {}, series: {}".format(ppa, series)

    if not series or not ppa:
        logging.error(
            "Missing compulsory environment variables (PPA, SERIES) "
            "{}".format(instance_info))
        return 1

    # Prepare launchpad connection:
    lp_series = launchpadmanager.get_series(series, distribution)
    monitored_ppa = launchpadmanager.get_ppa(ppa)
    # getting the archives and series objects
    if env.DEST:
        dest_archive = launchpadmanager.get_ppa(env.DEST)
    else:
        dest_archive = launchpadmanager.get_distribution(
            distribution).main_archive

    (
        available_archs_in_ppa,
        arch_all_arch,
        archs_to_eventually_ignore,
        archs_to_unconditionally_ignore,
    ) = launchpadmanager.get_available_all_and_ignored_archs(
        lp_series,
        monitored_ppa)

    # Collecting all packages that have been uploaded to the ppa
    packages_not_in_ppa = set()
    packages_building = set()
    packages_failed = set()
    packages = packageinppamanager.get_all_packages_uploaded()
    for (source, version, tip_rev, target_branch) in packages:
        packages_not_in_ppa.add(
            PackageInPPA(
                source,
                version,
                monitored_ppa,
                dest_archive,
                lp_series,
                available_archs_in_ppa,
                arch_all_arch,
                archs_to_eventually_ignore,
                archs_to_unconditionally_ignore))

    # Check the status regularly on all packages
    time_start_watching = time.time()
    while(packages_not_in_ppa or packages_building):
        logging.info("Checking the status for {}".format(
            list_packages_info_in_str(
                packages_not_in_ppa.union(packages_building))))
        packageinppamanager.update_all_packages_status(
            packages_not_in_ppa, packages_building, packages_failed, env.ARCH)

        # if we have some packages failing and no more build in progress, exit
        if packages_failed and not (packages_building and packages_not_in_ppa):
            logging.info("Some of the packages failed to build: {}".format(
                list_packages_info_in_str(packages_failed)))
            return 1

        # if we have no package building or failing and have wait for long
        # enough to have some package appearing in the ppa, exit
        if packages_not_in_ppa and not packages_building and (
                (time.time() - time_start_watching) >
                TIME_BEFORE_STOP_LOOKING_FOR_SOURCE_PUBLISH):
            logging.info(
                "Some source packages were never published in the ppa: "
                "{}".format(list_packages_info_in_str(packages_not_in_ppa)))
            return 1

        # exit if all is fine
        if not (packages_not_in_ppa or packages_building or packages_failed):
            break

        time.sleep(TIME_BETWEEN_PPA_CHECKS)
    return 0

if __name__ == '__main__':
    if '--help' in sys.argv or '-h' in sys.argv:
        print(__doc__)
    else:
        sys.exit(main())
